'''
 * this file is part of project
 * PIC4BBB
 * 
 * Program and control for dsPIC33EP / PIC24EP
 * with BeagleBone BLACK
 * M. 2017

 Free, open-source, fully featured demonstrative software.

THIS SOFTWARE IS PROVIDED,
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
PROVIDER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

----------------------------------------------------------------------------------
Intel HEX32 for Microchip - text file with format:
AABBBBCC(xxxx)DD
ALL ARE HEXADECIMAL DIGITS!
AA - number of Bytes in data
BBBB - lower 16 bit current address, HB:LB
DD - check sum: sum of all Bytes inverting bits and add 1, reject to 1B
(xxxx) - data ->

CC - code

CC = 1: this is end of file
CC = 0: this is data for the current memory address
here data is in groups of 8 digits:
(xx xx xx xx) - PIC memory cell is 24 bit, 00 MMB MB LB LLB (4 Bytes, most significant is 00)
the data in hex is: (xx xx xx xx) -> (LLB LB MB MMB 0) (little endian)

CC = 4: this is high 16 bit address
data is 16 bit field (xxxx) -> HighB:LowB
the high 16 bit address is for all next rows in file, until new row with CC = 4

the resulting current addr. = (high 16 bit addr. + low 16 bit addr)/2 (Microchip divide by 2),
because
EACH DATA (32 bit, 4B) is in even number address, i.e. current addr. increases by 2 after write
one 32 bit data (24 bit are the real code; the Config. uses only some bits in lower Byte, the
User ID are lower 16 bit... see device datasheet, Special Features)

----------------------------------------------------------------------------------

This is script to convert dsPIC33EP / PIC24EP compiled code *.hex into
special *.pbin code, useful to automatic programming of the PIC in BBB.
The *.pbin is much smaller than *.hex (about 10 times smaller!). In the same time,
it is more complex format, proper for extra security purposes.
So, this *.pbin file is very convenient for programming on fly the PIC in the BBB, via
Internet or some Wireless connection and so on.

Way of using - as command:

> python pymcephex32.py name.hex

At success, the name.pbin is generated in the current folder.
'''

import sys, pickle
from math import ceil

 
if (len(sys.argv) < 2) or (not sys.argv[1].endswith('.hex')):
    print('enter HEX file name')
    sys.exit(-1)

# 'U' option for automatic line end recognition - wors with Linux, Win, Mac
with open(sys.argv[1], 'rU') as f:
    ihex32 = f.readlines()

# Helper functions
def chsum(x):
    return ((~sum(x))+1) & 0xff

BinData = dict()
for L in ihex32:
    lenB = int(L[1:3], 16)
    addr = int(L[3:7], 16)
    code = int(L[7:9], 16)
    data = []
    if code == 0:
        if lenB//4 == 0: print('no data'); break
        for i in range(0, lenB*2, 8):
            # i is number of hex digits for bytes in lenB, 8 digits are 32 bit -> one data cell
            data.append(int(L[15+i:17+i]+L[13+i:15+i]+L[11+i:13+i]+L[9+i:11+i], 16))
    elif code == 4: data = int(L[9:11]+L[11:13], 16)
    cs = int(L[-3:], 16)
    if cs != chsum([int(L[i:i+2], 16) for i in range(1, len(L)-3, 2)]):
        print('check sum error'); break;
   
    if code == 4: aH = data    
    elif code == 1: print('end .hex parsing'); break;
    elif code == 0:
        # read data
        A = ((aH << 16) | addr)//2
        for i in range(0, len(data)):
            if A in BinData: print('address overlay error {}'.format(hex(A))); raise Exception
            BinData[A] = data[i]
            A += 2
    else: print('error code'); break;
    
# now BinData is dictionary
# addr: data, each 24 bit

# divide the three memory areas: program, config. and user ID

# each memory block with 24 bit data; only first cell is starting 24 bit address
pdata = [0] # addresses from 0 with step 2 (i.e. 0, 2, 4...)
addr = 0
k=0
while True:
    try:
        if addr < 0x200: pdata.append(BinData.pop(addr, 0xFFFFFF)) # fill any gap in unused ivt
        else: pdata.append(BinData.pop(addr))
        addr += 2
    except KeyError:
        break
to4 = len(pdata[1:]) % 4 # [0] is address
pdata.extend([0xFFFFFF]*(4-to4)) # to make packets by 4 cells
# (this is possible to overwritte the Config., but with '1', no unlocked, and
# at the end step - the correct Configs. will be written)

iddata = [0x800FF8] # user ID, 4 cells of 16 bit data (if not in .hex, program 0xFFFF)
iddata.append(BinData.pop(0x800FF8, 0xFFFF))
iddata.append(BinData.pop(0x800FFA, 0xFFFF))
iddata.append(BinData.pop(0x800FFC, 0xFFFF))
iddata.append(BinData.pop(0x800FFE, 0xFFFF))

cdata = [] # config. data (16 bit, 8 bit effectively 0xFFxx), start addr. determines automatically
caddrs = sorted(BinData.keys())

if len(caddrs) != 6: print('no all configs or some program memory gap'); sys.exit(-1)

cdata.append(caddrs[0])
for a in caddrs:
    cdata.append(BinData[a])

# now pdata, cdata and iddata has all programming memory in format
# first cell - start addr. and the rest are the memory content
# pdata is in packets by 4 cells of 24 bit codes
# cdata and iddata are in 16 bit, even number
#
# in BBB application:
# first pdata must be programmed / verified, after that iddata, at the end - cdata
# locking at the end

# prepare data ready for using in BBB
# -----------------------------------

# BBB pdata is in packets of 4 cells, in pages of max 749 such packets
# for details - see PRU code comments

# pdata (24 bit cells of type MSB:LSW) convert to -> format: start addr, 4 cells packets of type ->
#    // [0]: MSB1:MSB0 :LSW0
#    // [1]: LSW1
#    // [2]: MSB3:MSB2 :LSW2
#    // [3]: LSW3
p = [] # pages for pdata in upper format
# Type casting - for Python2 compatibility
for i in range(int(ceil((len(pdata)-1)/float(0xbb4)))):
    Baddr = i*0xbb4 # 749*4 = 0xbb4
    p.append([2*Baddr]) # add new page, the addresses are even
    # fill the page
    for i in range(0, 0xbb4, 4):
        try:
            d0 = ((pdata[Baddr+2+i] & 0xFF0000) << 8) | pdata[Baddr+1+i]
            d1 = pdata[Baddr+2+i] & 0xFFFF
            d2 = ((pdata[Baddr+4+i] & 0xFF0000) << 8) | pdata[Baddr+3+i]
            d3 = pdata[Baddr+4+i] & 0xFFFF
            p[-1].extend([d0, d1, d2, d3])
        except IndexError:
            break
# end markers
for i in range(len(p)): p[i][-1] |= 0xFFFF0000

# Write output file for Python 2 on BBB
with open(sys.argv[1][:-4] + '.pbin', 'wb') as f:
    pickle.dump((p, iddata, cdata), f, protocol=2)

print('end')

